<%!
def inherit(context):
  if context.get('chrome'):
    return 'master.html'
  return None
%>
<%inherit file="${inherit(context)}"/>

<%namespace name="tables" file="tables.html"/>

<link rel="alternate" type="application/atom+xml" title="New Updates" href="${request.route_url('updates_rss') + '?' + request.query_string}"/>

% if chrome:
<div class="container p-t-2">
  <div class="row">
    <div class="col-md-12">
      <h3>Updates <small>
        % if page == 1:
        <a href="${request.route_url('updates_rss') + '?' + request.query_string}">
          <span class="fa fa-rss"></span>
        </a>
        % endif
        </small>
      </h3>
      ${self.pager.render(page, pages)}
  % endif
      ${tables.updates(updates, display_user, display_request)}
  % if chrome:
      ${self.pager.render(page, pages)}
    </div>
  </div>
</div>
% endif
